<?php
// $Id: extra_elements.inc,v 1.1 2006/08/08 00:52:49 drumm Exp $

/**
 * @file
 * Extra form elements. Usage:
 *
 *   function hook_elements() {
 *     return include './'. drupal_get_path('module', 'yourmodule') .'/extra_elements.inc';
 *   }
 * 
 * Put a copy of this file in your contributed module's directory.
 */

// Make sure we don't include everything more than once.
if (defined('EXTRA_ELEMENTS_INCLUDED')) {
  return array();
}
else {
  define('EXTRA_ELEMENTS_INCLUDED', TRUE);

  /**
   * Rough stab at some quick documentation for this
   *
   * $form['table'] = array(
   *   '#type' => 'form_table',
   *   '#header' = array(t('the usual header array')),
   * );
   * $form['table'][row id] = array(
   *   '#tree' => TRUE
   *   //each form element child becomes a column
   *   'thing' => array(
   *     '#type' => 'textfield',
   *     // ...
   *   ),
   * );
   *
   * @todo Maybe automate the tree business.
   * @todo Some decent docs.
   * @todo Table cells as array('data' => ususal stuff, extra stuff) are
   *   impossible now.
   * @todo Implement everywhere needed in core.
   */
  function theme_form_table($element) {
    $rows = array();
    foreach (element_children($element) as $row_id) {
      $row = array();
      foreach (element_children($element[$row_id]) as $col_id) {
        unset($element[$row_id][$col_id]['#printed']);
        $row[] = form_render($element[$row_id][$col_id]);
      }
      $rows[] = $row;
    }
    return theme('table', $element['#header'], $rows);
  }

  function theme_multipage($element) {
    return $element['#children'];
  }

  /**
   * Implementation of hook_form_alter(). Here, we set up the 'page' field,
   * which keeps track of what stage the form is in.
   */
  /*
  function multipage_form_example_form_alter($form_id, &$form) {
    multipage_do_alter($form, $form);

    return $form;
  }
  */

  /**
   * Recursively search through the form tree for any multipage element. If one
   * is found, make a new hidden element to store the page number and do the
   * pre-render step.
   *
   * @param $form
   *   Always contains the entire form array being processed.
   * @param $element
   *   The current element being processed. Should be called with the root
   *   element.
   */
  function multipage_do_alter(&$form, &$element) {
    // Make sure it's our multipage form.
    if ($element['#type'] == 'multipage') {
      // Determine which page of the multipage form we're on. We don't do
      // any incrementing here - that's something that our #pre_render'er
      // will do when this page of the form has successfully validated.
      if (isset($_POST['edit']['page'])) {
        $page = $_POST['edit']['page'];
      }
      else if (isset($_GET['page'])) {
        $page = $_GET['page'];
      }
      else {
        $page = 1;
      }
      $element['page'] = array(
        '#type' => 'hidden',
        '#value' => $page,
      );

      // If back button is pressed, back up the form stage, Also, if preview
      // is hit we need to decrement the counter here to keep us on the same page.
      if ($_POST['op'] == $form['back']['#value']) {
        $element['page']['#value']--;
      }

      // Helper for our multipage - does field switching, sets options
      // based on validated $form_values instead of $_POST, and so forth.
      $form['#pre_render'] = array('multipage_pre_render');

      $links = array();
      foreach (element_children($element) as $page) {
        if (is_numeric($page)) {
          $links[] = l($element[$page]['#title'], $_GET['q'], array(), 'page='. $page);
        }
      }
      $form['page_links'] = array(
        '#type' => 'markup',
        '#value' => implode(' | ', $links),
        '#weight' => -100,
      );

      // This modifies the form for validation purposes. once validation is
      // completed, it'll be called one more time (through Drupal's Form API)
      // at which point it'll advance the form to the next page.
      multipage_pre_render($form_id, $form, FALSE);
    }
    else {
      foreach (element_children($element) as $key) {
        multipage_do_alter($form, $element[$key]);
      }
    }
  }

  /**
   * The #pre_render of a form allows us to make changes AFTER validation (unlike
   * hook_form_alter()), but BEFORE the form has actually been displayed. We use
   * it to control which form elements are shown, which are hidden, and which values
   * to set based on validate elements, not $_POST. This is a necessity for our
   * complicated multipage example form.
   */
  function multipage_pre_render($form_id, &$form, $next_page = TRUE) {
    multipage_do_pre_render($form, $form, $next_page);
  }

  /**
   * Recursively search through the form tree for any multipage element. If one
   * is found, manipulate the page number as needed and change the visiblility of
   * elements and title of the submit button.
   *
   * @param $form
   *   Always contains the entire form array being processed.
   * @param $element
   *   The current element being processed. Should be called with the root
   *   element.
   * @param $next_page
   *   FALSE if called during the alter step for initial processing; TRUE if
   *   called by form API for final processing.
   */
  function multipage_do_pre_render(&$form, &$element, $next_page) {
    if ($element['#type'] == 'multipage') {
      $page_count = max(element_children($element));
      // Are we ready for the next page in our form?
      if ($next_page && isset($_POST['edit']['page']) && ($_POST['op'] != $form['back']['#value'])) {
        $element['page']['#value'] += 1;
      }

      // Validation errors? Show previous page for correcting.
      if (form_get_errors()) {
        $element['page']['#value'] -= 1;
      }

      // If we have gotten to a page number that is higher than the number of
      // pages, move the page number back. For example, the 'Preview' button will
      // do this.
      if ($element['page']['#value'] > $page_count) {
        $element['page']['#value'] = $page_count;
      }

      $page_num = $element['page']['#value'];

      // Modify the page visibilities. The arrays tell us the pages the changes
      // should take place in.
      foreach (element_children($element) as $page) {
        if (is_int($page)) {
          multipage_set_element_visibility($element[$page], ($page_num == $page), $next_page);
        }
      }

      // Buttons
      multipage_set_element_visibility($form['back'],    ($page_num > 1), $next_page);
      multipage_set_element_visibility($form['next'],    ($page_num < $page_count), $next_page);
      multipage_set_element_visibility($form['preview'], ($page_num == $page_count), $next_page);
      multipage_set_element_visibility($form['submit'],  ($page_num == $page_count), $next_page);

      // The button text actually helps determine if a form has actually been
      // submitted because the name is also the value of a clicked button. By
      // changing it for building, but not rendering, the form is not fully
      // sumbitted until we name it 'Submit', the usual value.
      if ($next_page && $page_num < $page_count) {
        $form['next']['#value'] = t('Next (%page)', array('%page' => $element[$page_num + 1]['#title']));
      }
    }
    else {
      foreach (element_children($element) as $key) {
        multipage_do_pre_render($form, $element[$key], $next_page);
      }
    }
  }

  /**
   * Set an element's visibility. Elements are gnerally changed to hidden
   * elements. Visibility may be set and reset any number of times.
   *
   * @param $element
   *   The form element array to modify.
   * @param $visible
   *   The desired visibity of the form element.
   * @param $next_page
   *   Boolean value, TRUE if the next page is about to be rendered, FALSE otherwise.
   */
  function multipage_set_element_visibility(&$element, $visible, $next_page = TRUE) {
    multipage_restore_attributes($element);
    if (!$visible) {
      switch ($element['#type']) {
        case 'textfield':
        case 'textarea':
          multipage_set_attribute($element, '#type', 'hidden');
          multipage_set_attribute($element, '#required', FALSE);
          break;

        case 'select':
          if ($element['#multiple'] == TRUE) {
            if ($next_page) {
              multipage_set_attribute($element, '#type', 'hidden_array');
              $element = form_builder('foo', $element);
            }
          }
          else {
            multipage_set_attribute($element, '#type', 'hidden');
          }
          multipage_set_attribute($element, '#required', FALSE);
          break;

        case 'radios':
          // Radios elements cannot be hidden unless they have been processed.
          if ($next_page) {
            multipage_set_attribute($element, '#type', 'hidden');
            multipage_set_attribute($element, '#required', FALSE);
          }
          break;

        case 'radio':
        case 'checkbox':
          // We can't change these to hidden until right before the next page is rendered, otherwise
          // the value is lost sometimes.
          if ($next_page) {
            multipage_set_attribute($element, '#type', 'hidden');
          }
          break;

        case 'fieldset':
          multipage_set_attribute($element, '#type', NULL);
          break;

        case 'form_table':
          if ($next_page) {
            multipage_set_attribute($element, '#type', NULL);
          }
          break;

        case 'markup':
          multipage_set_attribute($element, '#value', '');
          break;

        case 'button':
          if ($next_page) {
            multipage_set_attribute($element, '#type', 'value');
          }
          break;

        case 'submit':
          if ($next_page) {
            multipage_set_attribute($element, '#type', 'button');
          }
          break;
      }
    }

    foreach (element_children($element) as $key) {
      multipage_set_element_visibility($element[$key], $visible, $next_page);
    }
  }

  /**
   * Set an attribute on an element array with storing the previous value which
   * may be reverted to using multipage_restore_attributes().
   *
   * @param $element
   *   The form element array to modify.
   * @param $key
   *   A key of the form element array.
   * @param $new_value
   *   The new value for the attribute.
   */
  function multipage_set_attribute(&$element, $key, $new_value) {
    $element['#multipage_form_original_'. $key] = $element[$key];
    $element[$key] = $new_value;
  }

  /**
   * Restore any form attributes which have been set using
   * multipage_set_attribute().
   *
   * @param $element
   *   The form element array to restore.
   */
  function multipage_restore_attributes(&$element) {
    foreach (array_filter(array_keys($element), create_function('$key', 'return (strpos($key, "#multipage_form_original_") === 0);')) as $key) {
      $element[str_replace('#multipage_form_original_', '', $key)] = $element[$key];
    }
  }

  function expand_hidden_array($element) {
    $values = is_array($element['#value']) ? $element['#value'] : array();
    $element['#tree'] = TRUE;
    foreach ($values as $key => $value) {
      $element[$key] = array('#type' => 'hidden', '#title' => '', '#processed' => TRUE, '#value' => $value);
    }
    return $element;
  }

  function theme_hidden_array($element) {
    $values = is_array($element['#values']) ? $element['#values'] : array();
    return $element['#children'];
  }

  // Build the types array and return it as if we were in a function, except this
  // is an included file.
  $types = array();

  $types['form_table'] = array('#header' => array());
  $types['multipage'] = array();
  $types['hidden_array'] = array('#input' => TRUE, '#process' => array('expand_hidden_array' => array()), '#tree' => TRUE);

  return $types;
}
